import unittest
import sys
from pathlib import Path
PROJECT_DIR = str(Path(__file__).resolve().parents[1])
sys.path.insert(1, PROJECT_DIR)
from axelf_executable import Axelf


class AxelfExecutableTest(unittest.TestCase):
    def test_simple_example(self):
        peptides = [
            'ADMGHLKY',
            'ELDDTLKY',
            'FMDHVLRY',
            'FSDLPLRV',
            'FSDNIEFY',
            'FTELAILY',
            'GLDEPLLK',
            'GLDQPLLK',
            'HIDFAIQY'
        ]

        alleles = [
            'HLA-B*52:01',
            'HLA-C*12:02',
            'HLA-A*26:01',
            'HLA-B*38:01',
            'HLA-B*52:01',
            'HLA-C*12:02',
            'HLA-C*12:02',
            'HLA-C*12:03',
            'HLA-B*07:02',
            'HLA-B*07:02'
        ]

        expected_result = [(0.001882, 1.7530978245), (0.004367, 1.8749579876666667), (0.045223, 2.6271013698), (0.003321, 1.8170831563333334), (0.003313, 1.8166405190000001), (0.038183, 2.504744748375), (3.2e-05, 1.605311648), (6e-06, 1.6009959340000002), (0.000503, 1.683492467)]

        axelf_obj = Axelf()
        score = axelf_obj.calculate_ic50_score(peptides, alleles)
        self.assertListEqual(expected_result, score)


    def test_duplicate_peptides(self):
        peptides = [
            'ADMGHLKY',
            'ELDDTLKY',
            'ELDDTLKY',
            'FMDHVLRY',
            'FSDLPLRV',
            'FSDNIEFY',
            'FTELAILY',
            'GLDEPLLK',
            'GLDQPLLK',
            'HIDFAIQY'
        ]

        alleles = [
            'HLA-B*52:01',
            'HLA-C*12:02',
            'HLA-A*26:01',
            'HLA-B*38:01',
            'HLA-B*52:01',
            'HLA-C*12:02',
            'HLA-C*12:02',
            'HLA-C*12:03',
            'HLA-B*07:02',
            'HLA-B*07:02'
        ]

        expected_result = [(0.001882, 1.7530978245), (0.004367, 1.8749579876666667), (0.094254, 3.1640536764761906), (0.003159, 1.8081197503333333), (0.074521, 2.9747460128095238), (0.122469, 3.3722651082941173), (0.038183, 2.504744748375), (5e-06, 1.600829945), (3.4e-05, 1.605643626), (0.000503, 1.683492467)]

        axelf_obj = Axelf()
        score = axelf_obj.calculate_ic50_score(peptides, alleles)
        self.assertListEqual(expected_result, score)


if __name__ == '__main__':
    unittest.main()